--- Todo: turn this thing into widget handler that will load other lua files for doing different things
--[[
Functions of the engine:

LogMessage(...)
logs the message.
GetOption(string)
gets option, returns string or nothing if option does not exist. Todo: table for options
SetOption(string, value)
Sets the option.


Coding convention:

The functions with_underscope_names are called by engine.
Returning true from such function supresses engine's built-in handler, whenever that makes sense.
For example, sound_on_chaser_death is called by engine, and if you add return true to it, engine will not make sound by itself.
(chaser is the name of those things that run after player and shoot at him)

]]

main_loaded=true

function GetMyID()
	return 1
end

my_state={get_all=true}

objects={}


--- set t[k]=t[dk] if t[k] does not exist
function DefaultTableValue(t,k,dk)
	if t[dk] and not t[k] then
		t[k]=t[dk]
	end
end

NewGameObject_=NewGameObject

local my_defaults={
old_pos="pos",
old_orientation="orientation"
}

function NewGameObject(t)
	for k,dk in pairs(my_defaults) do
		DefaultTableValue(t,k,dk)
	end
	return NewGameObject_(t)
end



function table.val_to_str ( v )
  if "string" == type( v ) then
    v = string.gsub( v, "\n", "\\n" )
    if string.match( string.gsub(v,"[^'\"]",""), '^"+$' ) then
      return "'" .. v .. "'"
    end
    return '"' .. string.gsub(v,'"', '\\"' ) .. '"'
  else
    return "table" == type( v ) and table.tostring( v ) or tostring( v )
  end
end

function table.key_to_str ( k )
  if "string" == type( k ) and string.match( k, "^[_%a][_%a%d]*$" ) then
    return k
  else
    return "[" .. table.val_to_str( k ) .. "]"
  end
end

function table.tostring( tbl )
  local result, done = {}, {}
  for k, v in ipairs( tbl ) do
    table.insert( result, table.val_to_str( v ) )
    done[ k ] = true
  end
  for k, v in pairs( tbl ) do
    if not done[ k ] then
      table.insert( result,
        table.key_to_str( k ) .. "=" .. table.val_to_str( v ) )
    end
  end
  return "{\n" .. table.concat( result, "," ) .. "}\n"
end

function table.write_value(v, file, indent_depth)
	if "string" == type( v ) then
		v = string.gsub( v, "\n", "\\n" )
		if string.match( string.gsub(v,"[^'\"]",""), '^"+$' ) then
			file:write( "'" .. v .. "'")
		end
		file:write( '"' .. string.gsub(v,'"', '\\"' ) .. '"')
	elseif "table" == type( v ) then
		table.write(v, file, indent_depth)
	else
		file:write(tostring(v))
	end
end

function table.write(tbl, file, indent_depth)
  local indent_depth=indent_depth or 0.0;
  local result, done = {}, {}
  file:write("{\n")
  local need_comma=false
  for k, v in ipairs( tbl ) do
	if need_comma then
		file:write(",\n")
	end
	file:write(string.rep(" ",indent_depth+1))
	table.write_value(v, file, indent_depth+1)
    done[ k ] = true
    need_comma=true
  end
  for k, v in pairs( tbl ) do
    if not done[ k ] then
		if need_comma then
			file:write(", \n")
		end
		file:write(string.rep(" ",indent_depth+1))
		file:write(table.key_to_str(k))
		file:write("=")
		table.write_value(v, file, indent_depth+1)
		need_comma=true
    end
  end
  file:write("\n")
  file:write(string.rep(" ",indent_depth))
  file:write("}")
end

function on_object_created(id)
	--LogMessage("on_object_created")
	objects[id]={}
	--[[
	objects[id]={get_all=true}
	GetObjectState(id, objects[id])
	LogMessage(table.tostring(objects[id]));
	]]
end

function on_object_erased(id)
	--LogMessage("on_object_erased")
	objects[id]=nil
end

function on_object_death(id)
	--LogMessage("on_object_death")
end

function override_field_of_view(fov)
end



-- Following parameters are updated by the engine, and are listed here for convenience only
audio_visualizer_in={
amplitude=0,
smooth_amplitude=0,
beat_amplitude=0, -- equals amplitude/smooth_amplitude when smooth amplitude is greater than some threshold
}
-- Following parameters have to be set by scripts
audio_visualizer_out={
background_offset_x=2,-- shifting of background over time
background_offset_y=0,
background_update_speed=0,-- speed of updating of background, combines with frameskip
background_phase={0,0,0} -- use this for blinking. Shifts the colours. format is rgb
}
local avi=audio_visualizer_in --- shorter abbrevations (also, locals like that make script marginally faster)
local avo=audio_visualizer_out


function Len2(vector)
	local len2=0;
	for _,a in ipairs(vector) do
		len2=len2+a*a;
	end
	return len2
end

function Len(vector)
	local len2=0;
	for _,a in ipairs(vector) do
		len2=len2+a*a
	end
	return math.sqrt(len2)
end

function Normalize(vector)
	local len=0
	for _,a in ipairs(vector) do
		len=len+a*a
	end
	len=1.0/math.sqrt(len)
	for i,a in ipairs(vector) do
		vector[i]=a*len
	end
end

function Normalized(vector)
	local len=0;
	for _,a in ipairs(vector) do
		len=len+a*a;
	end
	len=1.0/math.sqrt(len)
	result={}
	for i,a in ipairs(vector) do
		result[i]=a*len;
	end
	return result
end

function AxisAngleToQuaternion(axis, angle)
	local a=Normalized(axis)
	local s=math.sin(0.5*angle)
	local c=math.cos(0.5*angle)
	return {c,s*a[1],s*a[2],s*a[3]}
end

function AddVec(vector1,vector2)
	result={}
	for i,a in ipairs(vector1) do
		result[i]=vector1[i]+vector2[i]
	end
	return result
end

function MulVec(vector1,vector2)
	result={}
	for i,a in ipairs(vector1) do
		result[i]=vector1[i]*vector2[i]
	end
	return result
end

function SubVec(vector1,vector2)
	result={}
	for i,a in ipairs(vector1) do
		result[i]=vector1[i]-vector2[i]
	end
	return result
end

function ScaleVec(vector,scalar)
	result={}
	for i,a in ipairs(vector) do
		result[i]=vector[i]*scalar
	end
	return result
end

function MulQuatQuat(q1,q2)
	return {
	q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3] - q1[4]*q2[4],
	q1[1]*q2[2] + q1[2]*q2[1] + q1[3]*q2[4] - q1[4]*q2[3],
	q1[1]*q2[3] - q1[2]*q2[4] + q1[3]*q2[1] + q1[4]*q2[2],
	q1[1]*q2[4] + q1[2]*q2[3] - q1[3]*q2[2] + q1[4]*q2[1]
	}
end

function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

Transform={}
function Transform:new(o)
	o = o or {pos={0,0,0} , orientation={1,0,0,0}}
	setmetatable(o, self)
	self.__index=self
	return o;
end
function Transform:Translate(v)--- translate vector
	return Transform:new({pos=AddVec(self.pos,MulQuatVec(self.orientation,v)), orientation=self.orientation})
end
function Transform:Rotate(axis,angle)---
	return Transform:new({
		pos=self.pos,
		orientation=MulQuatQuat(self.orientation, AxisAngleToQuaternion(axis, angle*math.pi/180.0))
	})
end



-- By the way. Variables that you create should be local


--[[

local level_script = nil

local level_script_env = nil

-- Current global environment appears read only to the level script,
-- to minimize risk of script leaving some changes as levels are reloaded.
-- todo: in C++ , wipe our entire lua environment on level change?
local eproxy = {}
eproxy.__parent = getfenv(0)
eproxy.__index = function(t, n)
	local parent = rawget(getmetatable(t), "__parent")
	if env and rawget(t,n) then return rawget(env,n) end

	if not parent then return nil end
	return parent[n]
end
eproxy.__newindex = function(t, n, v)
	local parent = rawget(getmetatable(t), "__parent")
	-- if rawget(parent,n) then parent[n] = v end
	rawset(t,n,v)
end

function on_load_level(a)
	LogMessage("lua: on_load_level", a)
	if a and a~="" then
		local filename=VFSTranslate("$$data/levels/" .. a);
		local err
		level_script, err=loadfile(filename)
		if not level_script then
			LogMessage("Lua error loading ", filename, ":", err)
			return
		end
		level_script_env = {}
		setmetatable(level_script_env,eproxy)
		setfenv(level_script,level_script_env)
		local status, msg=pcall(level_script)
		if not status then
			LogMessage("Lua error running ", filename, ":", err)
		end
		for n,v in pairs(level_script_env) do
			LogMessage(n,"=",v)
		end
	else
		level_script_env = nil
		level_script = nil
	end
end
--]]

--- don't draw with opengl from on_ event handlers

local flowers_to_spawn=10
local spawn_countdown=1
local spawn_interval=5

Script={}

coroutine_data={}
--[[
function InitCoroutine(f)
	table.insert(coroutine_threads,coroutine.create(f));
end

function InitScript()
	for k,v in pairs(Script) do
		coroutine_data
	end
end]]

function RunScript()
	for k,v in pairs(Script) do
		local data=coroutine_data[k]
		if not data then
			data={cr=coroutine.create(v)}
			coroutine_data[k]=data
			--LogMessage("Script '" .. k .. "' started")
		end
		if data.cr then
			if coroutine.status(data.cr) =="dead" then
				data.cr=nil
				LogMessage("Script " .. k .. " completed")
			else
				local success,msg=coroutine.resume(data.cr)
				if not success then
					data.cr=nil
					LogMessage("Script " .. k .. " completed with message "..msg)
				end
			end
		end
	end
end

WaitNextFrame=coroutine.yield

game_time=0
game_dt=0

function Sleep(time)
	local end_time=game_time+(time or 0)
	repeat
		WaitNextFrame()
	until game_time>=end_time
end

function on_game_start()
	--InitScript()
	return true
end

function on_game_frame(dt) --- called by engine
	os.setlocale("C")-- fixing some weird bug
	game_dt=dt
	game_time=game_time+dt
	GetObjectState(GetMyID(), my_state)
	RunScript()
end

function on_screen_frame(dt) --- called by engine
	--LogMessage("lua on_screen_frame",dt)
end

function on_pre_draw()
	--test_shit_fn0(1,2,3,4)
end

audio_x=0 --- used by drawer
notes=nil
note_smoothing_time=0.1
smooth_notes={};
smooth_avg_amp=0
for i=1,512 do -- hack
		smooth_notes[i]=0
end

function on_audio_frame(dt) --- called by engine
	os.setlocale("C")-- fixing some weird bug
	-- spams the console
	--LogMessage("lua on_audio_frame", audio_visualizer_in.amplitude, audio_visualizer_in.smooth_amplitude, audio_visualizer_in.beat_amplitude)
	local a=math.log(avi.amplitude*1E2 + 1)/math.log(1E2 + 1) --- maps amplitudes 0.01 .. 1 to 0..1 logarithmically
	local phase=a
	--avo.background_phase={phase,phase,phase}
	audio_x=audio_x+3

	notes=GetMusicNotes();
	if notes==nil then return end
	--- notes.steps_per_octave gives number of notes per octave (currently 48) ,
	--- notes.A440 gives index of note A with frequency 440 hz
	--- notes start from A subsubcontra and go up all way to about 22khz .

	local a=math.exp(-dt/note_smoothing_time)
	smooth_avg_amp=0;
	for i=1,#notes do
		smooth_notes[i]=smooth_notes[i]*a+notes[i]*(1-a)
		smooth_avg_amp=smooth_avg_amp+smooth_notes[i]
	end
	smooth_avg_amp=smooth_avg_amp/#notes
end
--[[
-- test evolver thing
evolver_data={}
local population=1000
local max_replicate=10
local mutate_amount=0.1
for i=1,population do
evolver_data[i]=SyncedRandom();
end

local function evolver_step()
	chosen={}
	for i=1,population do -- choose 20 items
		local index=SyncedRandom()*population
		while chosen[index] do
			index=SyncedRandom()*population
		end
		chosen[index]=evolver_data[index]
	end
	local min=1000.0, max=-1000.0;
	for n,v in pairs(chosen) do
		if v<min then
			min=v;
		end
		if v>max then
			max=v;
		end
	end
	if max>min then
		for n,v in pairs(chosen) do
			if v==min then --- winner replicates
				local rep=v*max_replicate
				for i=1,rep do
					index=1+SyncedRandom()*population
					local child_v=v+mutate_amount*(SyncedRandom()-0.5)
					if child_v<0 then child_v=0 end
					if child_v>1 then child_v=1 end
					evolver_data[index]=child_v
				end
			end
		end
	end
end
]]
--- draw with opengl from draw_ functions

DrawHUDFunc={}


function DrawHUDFunc.DrawMusicVisualizerDebug(width,height,bloom)
	if bloom then return false end

	if GetOption("devel/draw_debug_info")~="1" then return false end--- dont want this function now.

	if not notes then return end
	gl.Enable(gl.BLEND)
	gl.Begin(gl.QUADS);
	local piano_keyboard_height=20

	local black={false,true,false,true,false,false,true,false,true,false,true,false}

	local halftones=#notes*12/notes.steps_per_octave
	local halftone_width=notes.steps_per_octave/12

	for i=2,#notes-1 do
		local a=smooth_notes[i]
		local is_peak=a>0.52*(smooth_notes[i-1]+smooth_notes[i+1]) and a>smooth_avg_amp
		local x=i

		if is_peak then
			gl.Color(1,1,0)
		else
			gl.Color(0,1,0)
		end
		a=(math.log(a*1E5))*0.05
		if a>0 then
			--- vertices of a rectangle are specified like this
			gl.Vertex(x+width-#notes, height - piano_keyboard_height)
			gl.Vertex(x+width-#notes +1, height - piano_keyboard_height)
			gl.Vertex(x+width-#notes +1, height-a*height - piano_keyboard_height)
			gl.Vertex(x+width-#notes, height-a*height - piano_keyboard_height)
		end
	end
	--- draw ocrave markers
	local octaves=#notes/notes.steps_per_octave
	for i=0,octaves-1 do
		local x=i*notes.steps_per_octave+3*halftone_width
		gl.Color(0,0.1,0)
		gl.Vertex(x+width-#notes, height)
		gl.Vertex(x+width-#notes, height-300)
		gl.Color(0,0.2,0)
		gl.Vertex(x+width-#notes +notes.steps_per_octave, height - 300)
		gl.Vertex(x+width-#notes +notes.steps_per_octave, height)
	end

	--- draw 'keyboard' underneath
	for i=0,halftones-1 do
		local is_black=black[((i-3)%12) +1]
		if is_black then
			gl.Color(0,0.1,0)
		else
			gl.Color(0,0.3,0)
		end
		local x=i*halftone_width
		gl.Vertex(x+width-#notes, height )
		gl.Vertex(x+width-#notes +halftone_width-1, height)
		gl.Vertex(x+width-#notes +halftone_width-1, height - piano_keyboard_height)
		gl.Vertex(x+width-#notes, height - piano_keyboard_height)
	end

	gl.End();
end

function DrawHUDFunc.DrawCrosshair(width,height,bloom)
	if game_paused then
		return
	end
	if bloom then
		gl.Color(0,1,0)
	else
		gl.Color(1,1,0)
	end
	gl.Begin(gl.LINES)
	local x=width*0.5
	local y=height*0.5
	local s=20 -- cross hair size
	local gs=5 --gap size
	gl.Vertex(x-s,y)
	gl.Vertex(x-gs,y)

	gl.Vertex(x+s,y)
	gl.Vertex(x+gs,y)

	gl.Vertex(x,y-s)
	gl.Vertex(x,y-gs)

	gl.Vertex(x,y+s)
	gl.Vertex(x,y+gs)
	gl.End()
end

function draw_HUD(width,height,bloom)-- called by engine, return true to suppress built-in drawer.
	for k,v in pairs(DrawHUDFunc) do
		v(width,height,bloom)
	end
	return false -- do not override engine's DrawHUD
end

function draw_audio_onto_background(width,height)  --- called by engine, twice per frame now. return true to suppress built-in drawer.
	audio_x=audio_x%width
	--LogMessage("draw_audio_onto_background")
--[[
	gl.Begin(gl.QUADS);
	for i=2,#notes-1 do
		local a=smooth_notes[i]
		local is_peak=a>0.9*(smooth_notes[i-1]+smooth_notes[i+1]) and a>smooth_avg_amp and a<notes[i]
		local x=(i)/#notes

		if is_peak then
			local y0=0.5*(i-1)/#notes
			local y=0.5*(i)/#notes
			gl.Color(0,a,0)
			gl.Vertex((2.0*audio_x)/width-1.0, y0)
			gl.Vertex((2.0*audio_x)/width-1.0, y)

			gl.Color(a,0,0)
			gl.Vertex(2.0*(audio_x+4.0)/width-1.0, y)
			gl.Vertex(2.0*(audio_x+4.0)/width-1.0, y0)

			gl.Color(0,a,0)
			gl.Vertex((2.0*audio_x)/width-1.0, -y0)
			gl.Vertex((2.0*audio_x)/width-1.0, -y)

			gl.Color(a,0,0)
			gl.Vertex(2.0*(audio_x+4.0)/width-1.0, -y)
			gl.Vertex(2.0*(audio_x+4.0)/width-1.0, -y0)
		end
	end
	gl.End();
	]]

	gl.Begin(gl.QUADS);
	local a=avi.amplitude
	gl.Color(0,1,0)
	gl.Vertex((2.0*audio_x)/width-1.0,-a)

	gl.Color(0,1,0)
	gl.Vertex((2.0*audio_x)/width-1.0, -a/2)

	gl.Color(1,0,0)
	gl.Vertex(2.0*(audio_x+4.0)/width-1.0, -a/2)

	gl.Color(1,0,0)
	gl.Vertex(2.0*(audio_x+4.0)/width-1.0,-a)


	gl.Color(0,1,0)
	gl.Vertex((2.0*audio_x)/width-1.0,a)

	gl.Color(0,1,0)
	gl.Vertex((2.0*audio_x)/width-1.0, a/2)

	gl.Color(1,0,0)
	gl.Vertex(2.0*(audio_x+4.0)/width-1.0, a/2)

	gl.Color(1,0,0)
	gl.Vertex(2.0*(audio_x+4.0)/width-1.0,a)
	gl.End();


	return true-- if you return true, built-in drawer does not run.
end

--- px,py.pz is position of impact; vx,vy,vz is velocity of object; bvx,bvy,bvz is velocity of bullet.
function sound_on_chaser_hit(health, px,py,pz, vx,vy,vz, bvx,bvy,bvz)-- called by engine, return true to suppress built-in.

end

--legacy shit
function on_chaser_death(px,py,pz, vx,vy,vz)-- called by engine, return true to suppress built-in.
	return false
end

function on_chewer_death(px,py,pz, vx,vy,vz)-- called by engine, return true to suppress built-in.
	return false
end

local function NoteOffsetToPitch(i)-- i: offset in half tones. returns: the pitch value for given offset. you need to multiply pitch values
	return 2.0^(i/12.0)
end

function sound_on_chaser_death(px,py,pz, vx,vy,vz)-- called by engine, return true to suppress built-in.
	so={pos={px,py,pz}, file="$$data/click1.wav", lifetime=0.5, vel={vx,vy,vz}, sound_gain=1, sound_pitch=0.25+2.0*math.random()}
	LogMessage("not creating sound from lua")
	NewSound(so)
	return true
end


function on_object_collision(id1,id2)
	--LogMessage("on_object_collision", id1, id2)
end

function on_add_score(score, pos)
	local s=""
	local color={1,1,1}
	if score>0 then
		s="+"..score
		color={0.2,1,0.04}
	else
		s=score
		color={1,0.2,0.04}
	end
	local q=my_state.orientation
	local marker_pos=AddVec(pos, MulQuatVec(q, {0,0.5,0}))
	NewGameObject({
	type="TextBillboard",
	text=s, lifetime=6, fadetime=2, color=color, pos=marker_pos, scale=0.05})
end

function gl.SetUniforms(s,values)
	gl.GetError()
	for k,v in pairs(values) do
		local loc=gl.GetUniformLocation(s,k)
		local err=gl.GetError()
		if err=="NO_ERROR" then
			gl.Uniform(loc, v)
			local err=gl.GetError()
			if err ~="NO_ERROR" then
				LogMessage("LuaGL error for "..k.." after Uniform:" .. err)
			end
		else
			LogMessage("LuaGL error for "..k.." after GetUniformLocation:" .. err)
		end
	end
end

function load_scenefractal_shader(brightness, time, low, high)

	local s=GetShader("$$data/test_v.glsl", "$$data/test_f.glsl")
	gl.UseProgram(s)
	gl.SetUniforms(s, {brightness=brightness})
end
--[[
function Script:SpawnLevelPortal()
	Sleep(GetOption("game/level_change_seconds")+0)
	--local my_state={}
	--GetObjectState(GetMyID(), my_state)
	--LogMessage("my_state=", table.tostring(my_state))
	local q=my_state.orientation
	local portal_pos=AddVec(my_state.pos, MulQuatVec(q, {0,0,-20}))
	NewGameObject{
	type="LevelPortal",
	pos=portal_pos,
	orientation=MulQuatQuat(my_state.orientation,AxisAngleToQuaternion({1,0,0},SyncedRandom()*math.pi/8)),
	level=GetOption("game/arena")+1,
	fade_in_time=5}
end
]]

function Script:SpawnLevelPortals()
	--local my_state={}
	--GetObjectState(GetMyID(), my_state)
	--LogMessage("my_state=", table.tostring(my_state))
	Sleep(1)
	NewGameObject{
	type="LevelPortal",
	pos={20,0,0},
	level=GetOption("game/arena")+1,
	fade_in_time=5,
	size=2
	}
	Sleep(1)
	NewGameObject{
	type="LevelPortal2",
	pos={20,20,0},
	visualizer=GetOption("game/music_visualizer")+1,
	fade_in_time=5,
	size=2
	}
end

--[[
function Script:SpawnLevelPortalExit()
	Sleep()
	NewGameObject{type="LevelPortalExit", pos=my_state.pos, orientation=my_state.orientation}
end
]]

--[[

static const float initial_chaser_spawn_rate=0.30;
//static const float initial_chaser_spawn_rate=2.0;///test


static const float chaser_spawn_rate_coefficient=5E-4;/// growth of chaser spawn rate with score
//static const float max_chaser_spawn_rate=2.0;///
static const float max_chaser_spawn_rate=2000.0;
static const float max_chaser_health=300.0;

static const float chaser_death_bonus_spawn_rate=0.3;

]]

function Script:SpawnFlowers()
	for i=1,10 do
		Sleep(spawn_interval)
		local id=NewGameObject{type="Flower", pos={RandomPoint()}, poly_n=SyncedRandom()*5, size=SyncedRandom()*10.0 }
		local test_data={get_all=true}
		GetObjectState(id,test_data)
		LogMessage(table.tostring(test_data));
	end
end

function Script:SpawnBots()
	while true do
		Sleep()
		local difficulty=GetOption("game/difficulty")+0;
		local max_objects=GetOption("visual/max_game_objects")+0;
		if #objects < max_objects then
			local spawn_rate=(0.1 + 0.0001 *my_state.score)*(0.5+difficulty);
			if SyncedRandom()<spawn_rate*game_dt then --- todo: honor max game objects?
				if SyncedRandom()<0.5 and difficulty>0 then
					NewGameObject{type="Chewer", pos={RandomPoint()}}
				else
					NewGameObject{type="Fighter", pos={RandomPoint()}}
				end
			end
		end
	end
end

local difficulty=GetOption("game/difficulty")+0;
if difficulty>0 then
	function Script:SpawnStartBonus()
		local pos=my_state.pos
		Sleep(0.5)
		NewGameObject{type="SimpleBonusPod", pos={pos[1]+20,pos[2], pos[3]}, bonus_id=0}
		Sleep(0.5)
		NewGameObject{type="SimpleBonusPod", pos={pos[1]+20,pos[2]+10, pos[3]}, bonus_id=1}
		Sleep(0.5)
		NewGameObject{type="SimpleBonusPod", pos={pos[1]+20,pos[2]+20, pos[3]}, bonus_id=2}
	end
end

function Script:SpawnMeteorites()
--[[
	while true do
		Sleep(-math.log(SyncedRandom()+1E-9)*5.0)
		LogMessage("Spawning meteorites")
		NewGameObject{
			type="Meteorites",
			pos={100*(SyncedRandom()-0.5), 300*(SyncedRandom()-0.5),-150},
			vel={0,0,10}
		}
	end
	]]
	while true do
		Sleep(0.5)
		local p={RandomPoint()}
		NewGameObject{
			type="Meteorites",
			pos={p[1], p[2], -100},
			vel={0,0,10},
			lifetime=100,
			mass=-math.log(SyncedRandom()+1E-9)*3.0
		}
	end
end

function sound_on_chewer_death()
	return false
end

function LocalToGlobalPos(p)
	return AddVec(my_state.pos, MulQuatVec(my_state.orientation, p))
end

billboard_opts={
		type="TextBillboard",	align_to_view=false, text="You forgot billboard text", lifetime=5,
		fadetime=1, color={0.3,0.4,0.5}, scale=0.1
	}

billboard_rescale=1

if(_T("$$lang")=="jp") then
	billboard_opts.font="$$data/sans.ftli"
	billboard_opts.font_size=30
	billboard_rescale=16/billboard_opts.font_size
	do_japanese_hacks=true
end

function NewBillboard(t)
	LogMessage("NewBillboard called")
	local q=my_state.orientation
	local marker_pos=AddVec(my_state.pos, MulQuatVec(q, t.offset or {0,0,-30}))
	LogMessage("NewBillboard calculations done")
	local t2={
		orientation=q, pos=marker_pos, scale=0.1
	}
	for k,v in pairs(billboard_opts) do
		t2[k] = v
	end

	LogMessage("NewBillboard table created")
	for k,v in pairs(t) do
		t2[k] = v
	end

	t2.scale=t2.scale*billboard_rescale

	LogMessage("NewBillboard table updated")
	LogMessage("about to call NewGameObject with following params" .. table.tostring(t2))
	local id=NewGameObject(t2)
	if t.audio then
		so={pos=marker_pos, file=t.audio, lifetime=1000, vel=my_state.vel, sound_gain=1, sound_pitch=1}
		NewSound(so)
	end

	return id
end

local buff=""
local max_buff=10
local function BufferMatch(s)
	return (s==buff:sub(-#s))
end

function on_char(c)
	--LogMessage("on_char",c)
	buff=buff .. c
	if #buff > max_buff then
		buff=buff:sub(1+#buff-max_buff)
	end

	if BufferMatch("hello") then
		local q=my_state.orientation
		local p=my_state.pos
		function PtTransform(a)
			LogMessage("pttransform")
			return AddVec(p, MulQuatVec(q, AddVec(a,{0,0,-120})))
		end
		function VTransform(a)
			return MulQuatVec(q, a)
		end
		require("levels.ilu")
	end

	if BufferMatch("snow") then
		require("levels.special_snowflakes_spawn")
	end


	if c=="1" then
		NewBillboard{text="The Polynomial", offset={0,0,-30}}
	elseif c=="2" then
		NewBillboard{text="Space of the music", offset={3,-2,-30}}
	elseif c=="3" then
		NewBillboard{text="There are multiple... KITTEN!", offset={0,1,-50}, audio="$$data/kitten.ogg"}
	end

end

function on_key(key, action)
end


--[[
Option<gu::KeyBinding>  roll_left_key(KeyBinding("A"), "input/keys/roll_left");
Option<gu::KeyBinding> roll_right_key(KeyBinding("D"), "input/keys/roll_right");
Option<gu::KeyBinding> accelerate_key(KeyBinding("W"), "input/keys/accelerate");
Option<gu::KeyBinding> decelerate_key(KeyBinding("S"), "input/keys/decelerate");
Option<gu::KeyBinding> shoot_key(KeyBinding("Space"), "input/keys/shoot");
Option<gu::KeyBinding> zoom_key(KeyBinding("Left_Shift"), "input/keys/zoom");
Option<gu::KeyBinding> hide_HUD_key(KeyBinding("H"), "input/keys/hide_HUD");
Option<gu::KeyBinding> hide_radar_key(KeyBinding("R"), "input/keys/hide_radar");
Option<gu::KeyBinding> pause_key(KeyBinding("P"), "input/keys/pause");
Option<gu::KeyBinding> next_song_key(KeyBinding("."), "input/keys/next_song");
Option<gu::KeyBinding> prev_song_key(KeyBinding(","), "input/keys/prev_song");
Option<gu::KeyBinding> restart_song_key(KeyBinding("/"), "input/keys/restart_song");

]]


local initial_controller={}
function ControllerHintTest(h)
	return math.abs( initial_controller[h.ctrl]-controller[h.ctrl] )>0.1
end


hints={
{
ctrl="throttle",
text=string.format(_T("Scroll wheel and %s,%s keys\ncontrol throttle"),GetOption("input/keys/accelerate"),GetOption("input/keys/decelerate")),
audio="$$data/control_throttle.ogg",
f=ControllerHintTest
},
{
ctrl="roll",
text=string.format(_T("%s,%s rolls"),GetOption("input/keys/roll_left"),GetOption("input/keys/roll_right")),
audio="$$data/roll.ogg",
f=ControllerHintTest
},
{
ctrl="zoom",
text=string.format(_T("Right mouse button or\n%s zooms"), GetOption("input/keys/zoom")),
audio="$$data/zoom.ogg",
f=ControllerHintTest
}
}
function NewHint(t)
	table.insert(hints,t)
end

local function table_copy(t)
	local result={}
	for k,v in pairs(t) do
		result[k]=v
	end
	return result
end

function Script:UpdateHints()
	Sleep()
	for k,v in pairs(hints) do
		if not v.f then
			hints[k]=nil
		else
			if GetOption("hints/" .. (v.name or v.ctrl)) then
				hints[k]=nil
			end
		end
	end

	initial_controller=table_copy(controller)
	local function test_change()
		for k,v in pairs(hints) do
			if (not v.f) or v.f(v) then
				LogMessage("Hint not needed: "..v.text)
				hints[k]=nil
				SetOption("hints/"..(v.name or v.ctrl), "no")
			end
		end
	end
	while next(hints) do
		test_change()
		Sleep()
	end
end
function Script:ShowHints()
	Sleep(30)
	local interval=20
	repeat
		for k,v in pairs(hints) do
			if v.text then
				NewBillboard{text=v.text, offset={0,0,-30}, audio=v.audio}
				Sleep(interval)
			else
				hints[k]=nil
			end
		end
	until not next(hints)
end


--[[
if GetOption("hints/throttle")~="1" then
	function Script:TellScrollWheel()
		Sleep()
		local changed=false
		for i=1,3*70 do
			if controller.throttle<=0.9 then
				LogMessage("throttle touched")
				changed=true
			end
			Sleep()
		end
		if changed then
			SetOption("hints/throttle", "1")
		else
			NewBillboard{text="Scroll wheel and W S keys\ncontrol throttle", offset={0,0,-30}}
		end
	end
end
]]


--[[
if not IsFullVersion() then
	function Script:AskToBuy()
		local sleeptime=300
		while true do
			Sleep(sleeptime)
			NewBillboard{text="The Polynomial : Demo", offset={0,0,-30}}
			Sleep(5)
			NewBillboard{text="Please buy full version", offset={0,0,-30}}
			sleeptime=sleeptime*2
		end
	end
end
]]


LogMessage("main.lua loaded!")

LogMessage(table.tostring(CrossProdTest( {1,2,3}, {4,5,6} ) ))

-- testing:

--local tmp=MulQuatVec({1,2,3},{1,2,3,4})
--LogMessage(table.tostring(tmp))
